/************************************************************
	Alfred Lu	07/15/2007	Email:Alfred.L@hotmail.com

	IDE:
	MCU:	ATMega128
	SD Interface:	SPI Mode
************************************************************/
#include <avr/io.h>         // iom128.h
#include <avr/sfr_defs.h>
#include "main.h"
#include "SD.h"
#include "FAT.h"

// Write a BYTE to SPI
UINT8 SPI_RW_Byte(UINT8 dat)
{
	UINT8 temp = 0;
	SPDR = dat;
	while(!(SPSR & _BV(7)));
	temp = SPDR;
	return temp;
}

//
// Send a Command to SD card, all command are 6 bytes long
// Return: the second byte of response register of SD Card
// cmd : command code
// arg : address or parameter,If a command does not require an argument, the
//       value of this field should be set to zero
//
UINT8 SD_Send_Command(UINT8 cmd, UINT8 crc, UINT32 arg)
{
	UINT8 retry=0;
	UINT8 rt = 0;

	// wait card to release MISO line
	SPI_RW_Byte(0xFF);

	// assert cs signal low
    PORTB &= ~(_BV(0));
    PORTA &= ~(_BV(6));

	// Ncs period
	SPI_RW_Byte(0xFF);

	// 6 bytes command
	SPI_RW_Byte(cmd | 0x40); 	// build transmitter bit
	SPI_RW_Byte(arg>>24);		// MSB of arg first
	SPI_RW_Byte(arg>>16);
	SPI_RW_Byte(arg>>8);
	SPI_RW_Byte(arg & 0xFF); 	// LSB of arg last
	SPI_RW_Byte(crc);        	// no protected mode of CRC

	// wait 8 bit R1 response, only last 8 bit is used here.
    // no need to send Ncr 0xFF for we send it while wait for
	// response from MISO
	retry = 0;
	do
	{
		rt = SPI_RW_Byte(0xFF);
	}
	while ((rt == 0xFF) && (retry++<10));

	// Nec period, may not necessory
	SPI_RW_Byte(0xFF);

    // deassert cs signal high
    PORTB |= _BV(0);
	PORTA |= _BV(6);
    return(rt);
}

//
//	Software SD Reset
//  CMD0 CMD1 CMD59 CMD16
//
UINT8 SD_Reset()
{
	UINT8 i;
	UINT8 retry;
	UINT8 Rt = 0;

	//
	// poll 100 SPI clk to make sure, card insert finished
	//
	for (i=0; i<100; i++)
	{
		SPI_RW_Byte(0xFF);
	}

	//
	// The SD Memory Card wakes up in the SD mode.
    // It will enter SPI mode if the CS signal is asserted (negative)
    // during the reception of the reset command (CMD0) and the card is in idle_state.
    //
	retry = 0;
	do
	{
		Rt = SD_Send_Command(CMD0_GO_IDLE_STATE, 0x95, 0);
		if (retry++ > 10)
			return 1;
	} while( Rt != RES_R1_IN_IDLE_STATUS );

    //
    // a special command SD_SEND_OP_COND (CMD1) is designed to provide SD Memory
    // Card hosts with a mechanism to identify and reject cards which do not match
    // the VDD range desired by the host. This is accomplished by the host sending
    // the required VDD voltage window as the operand of this command. Cards which
    // can not perform data transfer in the specified range must discard themselves
    // from further bus operations and go into Inactive State.
    //
	retry = 0;
	do
	{
		delay_nus(100);
		Rt = SD_Send_Command(CMD1_SEND_OP_COND, 0xFF, 0);	// send active
		if (retry++ > 10)
			return 1;
	} while( Rt );

	while(SD_Send_Command(CMD59_CRC_ON_OFF, 0xFF, 0));		// Disable CRC
	while(SD_Send_Command(CMD16_SET_BLOCKLEN, 0xFF, 512));	// Set sector to be 512 (default block size is in CSD)
												// RW must not cross block boundary unless _BLK_MISALIGN
												// bits is set in CSD

	return 0;
}

//
// read one block
// sector: data address
// buffer: store read data
//
UINT8 SD_ReadSingleBlock(UINT32 sector, void* buffer)
{
	UINT8 rt;
	UINT16 i = 512;
	UINT8 retry=0;
	UINT8 *pt = buffer;

	do
	{
		rt = SD_Send_Command(CMD17_READ_SINGLE_BLOCK, 0xFF, sector<<9);
		if(retry++ > 10)
			return 1;
	} while(rt != 0x00);

    // assert cs signal low
	PORTB &= ~(_BV(0));
    PORTA &= ~(_BV(6));

	//wait to start receiving data
	while(SPI_RW_Byte(0xFF) != /*0xFF*/0xFE);   // Start bytes of data tokens

	//read 512 bytes
	do
	{
		*(pt++) = SPI_RW_Byte(0xFF);
	} while(--i>0);

	SPI_RW_Byte(0xFF);	// fate CRC
	SPI_RW_Byte(0xFF);

    // deassert cs signal high
    PORTB |= _BV(0);
	PORTA |= _BV(6);

	return 0;
}

//
// write one block
// sector: data address
// buffer: store read data
//
UINT8 SD_WriteSingleBlock(UINT32 sector, void* buffer)
{
	UINT8 rt;
	UINT16 i;
	UINT8 retry=0;
	UINT8 *pt = buffer;

	do
	{
		rt = SD_Send_Command(CMD24_WRITE_SINGLE_BLOCK, 0xFF, sector<<9); //write command
		if(retry++ > 10)
			return 1;
	} while(rt != 0x00);

    // assert cs signal low
	PORTB &= ~(_BV(0));
    PORTA &= ~(_BV(6));

	// send 6 Clock Impulse
	i = 6;
	do
	{
		SPI_RW_Byte(0xFF);
	} while(--i > 0);

	i = 512;
	SPI_RW_Byte(0xFE);	//send Start bytes of data tokens
	do
	{
		SPI_RW_Byte(*(pt++));
	} while(--i > 0);

	SPI_RW_Byte(0xFF);
	SPI_RW_Byte(0xFF);

	//judge if it successful
	rt = SPI_RW_Byte(0xFF);
	if( (rt & 0x1F) != RES_DATA_ACCEPTED )
	{
		// deassert cs signal high
        PORTB |= _BV(0);
    	PORTA |= _BV(6);
		return rt;
	}

	//wait no busy
	while(!SPI_RW_Byte(0xFF));

    // deassert cs signal high
    PORTB |= _BV(0);
	PORTA |= _BV(6);

	return 0;
}

//
//  memory capacity = BLOCKNR * BLOCK_LEN
//	BLOCKNR = (C_SIZE + 1)* MULT
//  MULT = exp(2,C_SIZE_MULT+2)
//  BLOCK_LEN = exp(2, READ_BL_LEN)
//
UINT32 SD_ReadCapacity()
{
	UINT8 rt;
	UINT16 i;
	UINT16 temp;
	UINT8 buffer[16];
	UINT32 Capacity;

	rt = SD_Send_Command(CMD9_SEND_CSD, 0xFF, 0); //READ CSD
	if(rt != 0x00)
		return rt;

    // assert cs signal low
	PORTB &= ~(_BV(0));
    PORTA &= ~(_BV(6));

	while(SPI_RW_Byte(0xFF) != 0xFE);

	// card response 16 bytes
	i = 16;
	do
	{
		buffer[16-i] = SPI_RW_Byte(0xFF);
	} while(--i > 0);

	SPI_RW_Byte(0xFF);
	SPI_RW_Byte(0xFF);

	SPI_RW_Byte(0xFF);

    // deassert cs signal high
    PORTB |= _BV(0);
	PORTA |= _BV(6);

	// C_SIZE
	i = buffer[6] & 0x03;
	i <<= 8;
	i += buffer[7];
	i <<= 2;
	i += ((buffer[8] & 0xC0) >> 6);

	// C_SIZE_MULT
	rt = buffer[9] & 0x03;
	rt <<= 2;
	rt += ((buffer[10] & 0x80) >> 7);

	// BLOCKNR
	rt += 2;
	temp = 1;
	while(rt)
	{
		temp *= 2;
		rt--;
	}

	Capacity = ((UINT32)(i+1))*((UINT32)temp);

	// READ_BL_LEN
	i = buffer[6] & 0x0F;

    // BLOCK_LEN
	temp = 1;
	while(i)
	{
		temp *= 2;
		i--;
	}

	// The final result
	Capacity *= (UINT32)temp;
	return Capacity;
}
